@@ -1,6 +1,7 @@
 /*
   Copyright Ⓒ 1997, 1998, 1999, 2000, 2001  joost witteveen
   Copyright Ⓒ 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009  Clint Adams
+  Copyright Ⓒ 2012 Mikhail Gusarov
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -39,13 +40,21 @@
 */
 #define _DARWIN_NO_64_BIT_INODE
 
+/* The helper _unix2003 version of this file calls a few functions in this file
+   that are marked with static_nonapple so that needs to become private instead
+*/
+#define static_nonapple __attribute__((visibility("hidden")))
+
 #ifndef __LP64__
 /*
    This file is for 32-bit symbols which do not have the "$UNIX2003" version.
 */
 #define _NONSTD_SOURCE
 #endif
-#endif
+#else /* !__APPLE__ */
+/* static_nonapple needs to become static in this case */
+#define static_nonapple static
+#endif /* !__APPLE__ */
 
 #include "config.h"
 #include "communicate.h"
@@ -100,6 +109,20 @@
 #if HAVE_FTS_H
 #include <fts.h>
 #endif /* HAVE_FTS_H */
+#ifdef __sun
+#include <sys/systeminfo.h>
+#endif
+#ifdef __APPLE__
+#include <paths.h>
+#include <sys/stat.h>
+#include <crt_externs.h>
+#if MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_5
+#include <spawn.h>
+#endif
+#ifndef va_copy
+#define va_copy(dest,src) ((dest)=(src))
+#endif
+#endif /* __APPLE__ */
 
 #if !HAVE_DECL_SETENV
 extern int setenv (const char *name, const char *value, int replace);
@@ -477,7 +500,7 @@ static int set_faked_egid(gid_t egid) {
   return 0;
 }
 
-static int set_faked_reuid(uid_t ruid, uid_t euid) {
+static_nonapple int set_faked_reuid(uid_t ruid, uid_t euid) {
   read_uids();
   if (ruid != (uid_t)-1 || euid != (uid_t)-1)
     faked_saved_uid = faked_effective_uid;
@@ -489,7 +512,7 @@ static int set_faked_reuid(uid_t ruid, uid_t euid) {
   return write_uids();
 }
 
-static int set_faked_regid(gid_t rgid, gid_t egid) {
+static_nonapple int set_faked_regid(gid_t rgid, gid_t egid) {
   read_gids();
   if (rgid != (gid_t)-1 || egid != (gid_t)-1)
     faked_saved_gid = faked_effective_gid;
@@ -546,7 +569,7 @@ static gid_t set_faked_fsgid(gid_t fsgid) {
 #endif
 
 
-static int dont_try_chown(){
+static_nonapple int dont_try_chown(){
   static int inited=0;
   static int donttry;
 
@@ -1528,7 +1551,7 @@ int fakeroot_isdisabled(void)
   return fakeroot_disabled;
 }
 
-#ifdef HAVE_SYS_ACL_H
+#ifdef HAVE_ACL_T
 int acl_set_fd(int fd, acl_t acl) {
   errno = ENOTSUP;
   return -1;
@@ -1550,6 +1573,11 @@ FTSENT *fts_read(FTS *ftsp) {
   }
 #endif /* LIBFAKEROOT_DEBUGGING */
   r=next_fts_read(ftsp);
+#ifdef __APPLE__
+  if (r && ((ftsp->fts_options & FTS_NOSTAT)
+            || r->fts_info == FTS_NS || r->fts_info == FTS_NSOK))
+    r->fts_statp = NULL;  /* Otherwise fts_statp may be a random pointer */
+#endif
   if(r && r->fts_statp) {  /* Should we bother checking fts_info here? */
 # if defined(STAT64_SUPPORT) && !defined(__APPLE__)
     SEND_GET_STAT64(r->fts_statp, _STAT_VER);
@@ -1610,14 +1638,14 @@ getattrlist(const char *path, void *attrList, void *attrBuf,
     return r;
   }
   if (options & FSOPT_NOFOLLOW) {
-    r=lstat(path, &st);
+    r=WRAP_LSTAT(path, &st);
   } else {
-    r=stat(path, &st);
+    r=WRAP_STAT(path, &st);
   }
   if (r) {
     return r;
   }
-  patchattr(attrList, attrBuf, st.st_uid, st.st_gid);
+  patchattr(attrList, attrBuf, st.st_uid, st.st_gid, st.st_mode);
 
   return 0;
 }
@@ -1645,13 +1673,386 @@ fgetattrlist(int fd, void *attrList, void *attrBuf,
   if (r) {
     return r;
   }
-  r=fstat(fd, &st);
+  r=WRAP_FSTAT(fd, &st);
   if (r) {
     return r;
   }
-  patchattr(attrList, attrBuf, st.st_uid, st.st_gid);
+  patchattr(attrList, attrBuf, st.st_uid, st.st_gid, st.st_mode);
 
   return 0;
 }
 #endif /* if HAVE_FGETATTRLIST */
+
+/*
+ * Prevent loss of DYLD_INSERT_LIBRARIES, FAKEROOTKEY and FAKED_MODE
+ * environment variables during packaging (packagemaker otherwise clears them).
+ */
+
+static char *dylib_insert;
+static char *fakeroot_key;
+static char *faked_mode;
+static int spawnexec_in_progress;
+static pid_t spawnexec_pid;
+
+static int is_spawnexec_active()
+{
+  if (getpid() != spawnexec_pid) {
+    spawnexec_in_progress = 0;
+  }
+  return spawnexec_in_progress;
+}
+
+static void enter_spawnexec()
+{
+  (void)is_spawnexec_active();
+  if (++spawnexec_in_progress == 1)
+    spawnexec_pid = getpid();
+}
+
+static void exit_spawnexec()
+{
+  (void)is_spawnexec_active();
+  if (spawnexec_in_progress)
+    --spawnexec_in_progress;
+}
+
+__attribute__((__constructor__,__used__)) static void retrieve_dylib();
+static void retrieve_dylib()
+{
+  const char *dil = getenv("DYLD_INSERT_LIBRARIES");
+  const char *frk = fakeroot_key = getenv("FAKEROOTKEY");
+  const char *fdm = faked_mode = getenv("FAKED_MODE");
+  if (frk) {
+    char *savekey = (char *)malloc(12 + strlen(frk) + 1);
+    if (savekey) {
+      sprintf(savekey, "FAKEROOTKEY=%s", frk);
+      fakeroot_key = savekey;
+    }
+  }
+  if (fdm) {
+    char *savemode = (char *)malloc(11 + strlen(fdm) + 1);
+    if (savemode) {
+      sprintf(savemode, "FAKED_MODE=%s", fdm);
+      faked_mode = savemode;
+    }
+  }
+  if (dil) {
+    char *dupe = strdup(dil);
+    if (dupe) {
+      char *ptr = dupe;
+      do {
+        char *next = strchr(ptr, ':');
+        if (next)
+          *next++ = '\0';
+        if (strstr(ptr, "libfakeroot") && strstr(ptr, ".dylib")) {
+          dylib_insert = ptr;
+          return;
+        }
+        ptr = next;
+      } while (ptr);
+      free(dupe);
+    }
+  }
+}
+
+static char **get_exec_envp(char *const envp[], char **insenvp)
+{
+  int i, envcount = 0;
+  char **altenv = NULL;
+
+  *insenvp = NULL;
+  if (!envp || !dylib_insert || !fakeroot_key ||
+      getenv("FAKEROOT_TEST_PRELOAD"))
+    return (char **)envp;
+
+  while (envp[envcount]) {
+    ++envcount;
+  }
+  altenv = (char **)malloc((envcount + 4) * sizeof(char *));
+  if (!altenv)
+    return (char **)envp;
+  memcpy(altenv, envp, (envcount + 1) * sizeof(char *));
+
+  for (i=0; i<envcount; ++i) {
+    if (strncmp(altenv[i], "FAKEROOTKEY=", 12) == 0)
+      break;
+  }
+  if (i >= envcount) {
+    altenv[envcount] = fakeroot_key;
+    altenv[++envcount] = NULL;
+  }
+
+  for (i=0; i<envcount; ++i) {
+    if (strncmp(altenv[i], "FAKED_MODE=", 11) == 0)
+      break;
+  }
+  if (i >= envcount && faked_mode) {
+    altenv[envcount] = faked_mode;
+    altenv[++envcount] = NULL;
+  }
+
+  for (i=0; i<envcount; ++i) {
+    if (strncmp(altenv[i], "DYLD_INSERT_LIBRARIES=", 22) == 0) {
+      if (strstr(altenv[i], dylib_insert))
+        return altenv;
+      else {
+        *insenvp = (char *)malloc(strlen(altenv[i]) + strlen(dylib_insert) + 2);
+        if (!*insenvp) {
+          free(altenv);
+          return (char **)envp;
+        }
+        sprintf(*insenvp, "DYLD_INSERT_LIBRARIES=%s:%s", dylib_insert,
+          altenv[i]+22);
+        altenv[i] = *insenvp;
+        return altenv;
+      }
+    }
+  }
+  *insenvp = (char *)malloc(22 + strlen(dylib_insert) + 1);
+  if (!*insenvp) {
+    free(altenv);
+    return (char **)envp;
+  }
+  sprintf(*insenvp, "DYLD_INSERT_LIBRARIES=%s", dylib_insert);
+  altenv[envcount] = *insenvp;
+  altenv[envcount+1] = NULL;
+  return altenv;
+}
+
+int execve(const char *path, char *const argv[], char *const envp[])
+{
+  int result;
+  char **altenv;
+  char *insenv;
+
+  if (!path || !argv || !envp || is_spawnexec_active()) {
+    enter_spawnexec();
+    result = next_execve(path, argv, envp);
+    exit_spawnexec();
+    return result;
+  }
+
+  altenv = get_exec_envp(envp, &insenv);
+  enter_spawnexec();
+  result = next_execve(path, argv, altenv);
+  exit_spawnexec();
+  if (altenv != envp)
+    free(altenv);
+  if (insenv)
+    free(insenv);
+  return result;
+}
+
+/* funnel execl, execle, execlp, execv, execvp and execvP into execve */
+
+int execv(const char *path, char *const argv[])
+{
+  return execve(path, argv, *_NSGetEnviron());
+}
+
+static char *find_exec_path(const char *file, const char *search_path)
+{
+  char *path = (char *)file;
+  if (file && *file && !strchr(file, '/') && search_path && *search_path) {
+    size_t splen = strlen(search_path);
+    const char *search_end = search_path + splen;
+    const char *next;
+    char *buff = malloc(strlen(file) + splen + 2);
+    if (!buff)
+      return NULL;
+    for (; search_path < search_end; search_path = next) {
+      struct stat info;
+      size_t plen;
+      if ((next = strchr(search_path, ':')))
+        plen = next++ - search_path;
+      else
+        plen = (next = search_end) - search_path;
+      memcpy(buff, search_path, plen);
+      sprintf(buff + plen, "%s%s", plen ? "/" : "", file);
+      if (stat(buff, &info) || !S_ISREG(info.st_mode) || access(buff, X_OK))
+        continue;
+      path = buff;
+      break;
+    }
+    if (path == file) {
+      free(buff);
+      errno = ENOENT;
+      return NULL;
+    }
+  }
+  return path;
+}
+
+int execvP(const char *file, const char *search_path, char *const argv[])
+{
+  int result;
+  char *path = find_exec_path(file, search_path);
+  if (!path)
+    return -1;
+  result = execve(path, argv, *_NSGetEnviron());
+  if (path != file)
+    free(path);
+  return result;
+}
+
+int execvp(const char *file, char *const argv[])
+{
+  const char *path = getenv("PATH");
+  return execvP(file, path ? path : _PATH_DEFPATH, argv);
+}
+
+static char **collect_args(const char *arg0, va_list args, char ***envp)
+{
+  char **arglist;
+  size_t count = 0;
+  char **outp;
+
+  va_list args_copy;
+  va_copy(args_copy, args);
+  for (;;) {
+    char *ap = va_arg(args_copy, char *);
+    if (!ap)
+      break;
+    ++count;
+  }
+  va_end(args_copy);
+  arglist = (char **)malloc((count + 2) * sizeof(char *));
+  if (!arglist)
+    return NULL;
+  outp = arglist;
+  *outp++ = (char *)arg0;
+  for (;;) {
+    char *ap = va_arg(args, char *);
+    *outp++ = ap;
+    if (!ap)
+      break;
+  }
+  if (envp)
+    *envp = va_arg(args, char **);
+  return arglist;
+}
+
+int execl(const char *path, const char *arg0, ...)
+{
+  int result;
+  char **arglist;
+  va_list args;
+  va_start(args, arg0);
+  arglist = collect_args(arg0, args, NULL);
+  va_end(args);
+  if (!arglist)
+    return -1;
+  result = execve(path, arglist, *_NSGetEnviron());
+  free(arglist);
+  return result;
+}
+
+int execle(const char *path, const char *arg0, ...)
+{
+  int result;
+  char **arglist;
+  char **envp;
+  va_list args;
+  va_start(args, arg0);
+  arglist = collect_args(arg0, args, &envp);
+  va_end(args);
+  if (!arglist)
+    return -1;
+  result = execve(path, arglist, envp);
+  free(arglist);
+  return result;
+}
+
+int execlp(const char *file, const char *arg0, ...)
+{
+  int result;
+  char **arglist;
+  va_list args;
+  va_start(args, arg0);
+  arglist = collect_args(arg0, args, NULL);
+  va_end(args);
+  if (!arglist)
+    return -1;
+  result = execvp(file, arglist);
+  free(arglist);
+  return result;
+}
+
+#if MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_5
+
+/* handle posix_spawn and posix_spawnp separately */
+
+int posix_spawn(pid_t * __restrict pid, const char * __restrict path,
+                const posix_spawn_file_actions_t *file_actions,
+                const posix_spawnattr_t * __restrict attrp,
+                char *const argv[ __restrict], char *const envp[ __restrict])
+{
+  int result;
+  char **altenv;
+  char *insenv;
+
+  if (!path || !argv || !envp || is_spawnexec_active()) {
+    enter_spawnexec();
+    result = next_posix_spawn(pid, path, file_actions, attrp, argv, envp);
+    exit_spawnexec();
+    return result;
+  }
+
+  altenv = get_exec_envp(envp, &insenv);
+  enter_spawnexec();
+  result = next_posix_spawn(pid, path, file_actions, attrp, argv, altenv);
+  exit_spawnexec();
+  if (altenv != envp)
+    free(altenv);
+  if (insenv)
+    free(insenv);
+  return result;
+}
+
+int posix_spawnp(pid_t * __restrict pid, const char * __restrict file,
+                 const posix_spawn_file_actions_t *file_actions,
+                 const posix_spawnattr_t * __restrict attrp,
+                 char *const argv[ __restrict], char *const envp[ __restrict])
+{
+  int result;
+  const char *search_path = getenv("PATH");
+  char *path = find_exec_path(file, search_path ? search_path : _PATH_DEFPATH);
+  if (!path)
+    return errno;
+  result = posix_spawn(pid, path, file_actions, attrp, argv, envp);
+  if (path != file)
+    free(path);
+  return result;
+}
+#endif /* MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_5 */
 #endif /* ifdef __APPLE__ */
+
+#ifdef __sun
+/*
+ * Disable the runtime selection of binaries on Solaris: libfakeroot is (yet?)
+ * unable to cope with several bitnesses of binaries being run.
+ */
+int sysinfo(int command, char *buf, long count)
+{
+    if (command == SI_ISALIST)
+    {
+        /* do the evil trick */
+#ifdef sparc
+#ifdef __arch64__
+        strncpy(buf, "sparcv9 sparc", count - 1);
+        return sizeof("sparcv9 sparc");
+#else
+        strncpy(buf, "sparcv7 sparc", count - 1);
+        return sizeof("sparcv7 sparc");
+#endif
+#else
+        strncpy(buf, "i386 i86", count - 1);
+        return sizeof("i386 i86");
+#endif
+    }
+    else
+    {
+        return next_sysinfo(command, buf, count);
+    }
+}
+#endif
